@@ -2,11 +2,11 @@ __author__ = 'Altertech, https://www.altertech.com/'
 __copyright__ = 'Copyright (C) 2019 Altertech'
 __license__ = 'MIT'
 
-__version__ = '0.4.9'
+__version__ = '0.4.20'
 
 import rapidtables
 import neotermcolor
-import datetime
+import datetime, time
 
 from functools import partial
 from collections import OrderedDict
@@ -21,7 +21,7 @@ from finac.core import ResourceAlreadyExists
 # asset methods
 from finac.core import asset_create, asset_delete
 from finac.core import asset_set_rate, asset_rate
-from finac.core import asset_delete_rate
+from finac.core import asset_delete_rate, asset_rate_range
 
 from finac.core import asset_update
 from finac.core import asset_precision
@@ -52,12 +52,15 @@ from finac.core import account_list, account_list_summary
 from finac.core import purge, transaction_purge
 
 # caches
-from finac.core import preload
+from finac.core import preload, exec_query
 
 # plots
 from finac.plot import account_plot as plot
 from finac.plot import account_pie as pie
 
+# df
+from finac.df import df
+
 # tools
 from finac.core import parse_number, parse_date, get_version
 
@@ -73,6 +76,22 @@ rate = asset_rate
 stmt = account_statement_summary
 
 
+def query(q):
+    t_start = time.time()
+    result = list(exec_query(q))
+    t_spent = time.time() - t_start
+    ft = rapidtables.format_table(result, fmt=rapidtables.FORMAT_GENERATOR)
+    if ft:
+        h, tbl = ft
+        neotermcolor.cprint(h, '@finac:title')
+        neotermcolor.cprint('-' * len(h), '@finac:separator')
+        for t in tbl:
+            print(t)
+        neotermcolor.cprint('-' * len(h), '@finac:separator')
+    print(f'SELECT {len(result)}')
+    print(f'Time: {t_spent:.3f}s')
+
+
 def check_version(warn=False):
     core_version = get_version()
     if __version__ != core_version:
@@ -168,7 +187,8 @@ def ls(account=None,
                 day=1, hour=0, minute=0, second=0, microsecond=0).timestamp(),
             end=end,
             tag=tag,
-            pending=pending)
+            pending=pending,
+            datefmt=True)
         stmt = result['statement'].copy()
         acc_info = account_info(account=account)
         precision = asset_precision(asset=acc_info['asset'])
@@ -184,7 +204,7 @@ def ls(account=None,
                    rapidtables.ALIGN_LEFT, rapidtables.ALIGN_LEFT,
                    rapidtables.ALIGN_LEFT, rapidtables.ALIGN_LEFT,
                    rapidtables.ALIGN_LEFT))
-        rcur = acc_info['asset']
+        rcur = base.upper() if base else acc_info['asset']
         if ft:
             h, tbl = ft
             neotermcolor.cprint(h, '@finac:title')
@@ -209,7 +229,6 @@ def ls(account=None,
             pl = result['debit'] - result['credit']
             if base:
                 pl = pl * asset_rate(acc_info['asset'], base, date=end)
-                rcur = base.upper()
             neotermcolor.cprint('{} {}'.format(format_money(pl, precision),
                                                rcur),
                                 attrs='bold',
@@ -323,7 +342,8 @@ def lsa(asset=None, start=None, end=None):
                 start=start if start else datetime.datetime.today().replace(
                     day=1, hour=0, minute=0, second=0,
                     microsecond=0).timestamp(),
-                end=end):
+                end=end,
+                datefmt=True):
             row = OrderedDict()
             row['pair'] = '{}/{}'.format(r['asset_from'], r['asset_to'])
             row['date'] = r['date']
