@@ -2,16 +2,21 @@ __author__ = 'Altertech, https://www.altertech.com/'
 __copyright__ = 'Copyright (C) 2019 Altertech'
 __license__ = 'MIT'
 
-__version__ = '0.4.9'
+__version__ = '0.4.20'
 
 from sqlalchemy.exc import IntegrityError
 from cachetools import TTLCache
 from itertools import groupby
 from .currencies import currencies
+from types import SimpleNamespace
+from concurrent.futures import ThreadPoolExecutor
 
-rate_cache = TTLCache(maxsize=100, ttl=5)
+_cache = SimpleNamespace(rate=None, rate_list=None)
 
-asset_precision_cache = {}
+_CacheRateKeyError = KeyError
+_CacheRateListKeyError = KeyError
+
+_asset_precision_cache = {}
 
 restrict_assets_to_currencies = False
 
@@ -97,6 +102,7 @@ from collections import OrderedDict
 from functools import wraps
 
 from pyaltt2.crypto import gen_random_str
+from pyaltt2.lp import parse_func_str
 from pyaltt2.converters import val_to_boolean, parse_date, parse_number
 
 import threading
@@ -109,6 +115,7 @@ _db = SimpleNamespace(engine=None, redis_conn=None)
 
 config = SimpleNamespace(db=None,
                          db_pool_size=10,
+                         thread_pool_size=30,
                          keep_integrity=True,
                          lazy_exchange=True,
                          full_transaction_update=True,
@@ -125,7 +132,11 @@ config = SimpleNamespace(db=None,
                          redis_timeout=5,
                          redis_blocking_timeout=5,
                          restrict_deletion=None,
-                         date_format='%Y-%m-%d %H:%M:%S')
+                         date_format='%Y-%m-%d %H:%M:%S %Z',
+                         rate_cache_ttl=None,
+                         insecure=False)
+
+_d = SimpleNamespace()
 
 lock_purge = threading.Lock()
 lock_account_token = threading.Lock()
@@ -139,6 +150,11 @@ multiply_fields = {
 }
 
 
+def _format_ttlcache_key(d, ttl):
+    t = d.timestamp()
+    return t if t < time.time() - ttl else None
+
+
 def _multiply(i):
     return round(i * config.multiplier) if config.multiplier and i else i
 
@@ -156,20 +172,26 @@ def core_method(f):
 
     @wraps(f)
     def do(*args, **kwargs):
-        if config.api_uri is not None:
+        if config.api_uri is None:
+            return f(*args, **kwargs)
+        else:
             import requests
             import uuid
             import json
+            req_id = str(uuid.uuid4())
             payload = {
                 'jsonrpc': '2.0',
                 'method': f.__name__,
                 'params': kwargs,
-                'id': str(uuid.uuid4())
+                'id': req_id
             }
             if config.api_key is not None:
                 payload['params']['_k'] = config.api_key
             for i, a in enumerate(args):
                 payload['params'][argspec.args[i]] = a
+            logger.debug('API request {} {} {}'.format(req_id,
+                                                       payload['method'],
+                                                       payload['params']))
             r = requests.post(config.api_uri,
                               json=payload,
                               timeout=config.api_timeout)
@@ -181,9 +203,9 @@ def core_method(f):
             if 'error' in result:
                 raise _exceptions.get(result['error']['code'], RuntimeError)(
                     result['error'].get('message'))
+            logging.debug('API response {} {}'.format(result['id'],
+                                                      result['result']))
             return result['result']
-        else:
-            return f(*args, **kwargs)
 
     return do
 
@@ -200,14 +222,14 @@ def deletion_method(f):
     return do
 
 
-def format_date(d):
+def format_date(d, force=False):
     if d is not None:
         if config.date_format is None:
             return d
         elif isinstance(d, int):
             return datetime.datetime.strftime(
                 datetime.datetime.fromtimestamp(d), config.date_format)
-        elif isinstance(d, datetime.datetime):
+        elif isinstance(d, datetime.datetime) and not force:
             return d
         else:
             return datetime.datetime.strftime(
@@ -219,7 +241,7 @@ def preload():
     Preload static data
     """
     for a in _asset_precision():
-        asset_precision_cache[a['asset']] = a['precision']
+        _asset_precision_cache[a['asset']] = a['precision']
 
 
 def asset_precision(asset):
@@ -227,8 +249,8 @@ def asset_precision(asset):
     Get precision (digits after comma) for the asset
     Note: asset precision is cached, so process restart required if changed
     """
-    return asset_precision_cache[
-        asset] if asset in asset_precision_cache else _asset_precision(
+    return _asset_precision_cache[
+        asset] if asset in _asset_precision_cache else _asset_precision(
             asset=asset)
 
 
@@ -237,17 +259,101 @@ def get_version():
     return __version__
 
 
+@core_method
+def exec_query(q, _grafana=False):
+    """
+    Execute statement
+
+    EXPERIMENTAL, DON'T USE IN PRODUCTION
+
+    Currenly only function call statements are supported:
+
+        SELECT <function>([args, kwargs])
+
+        e.g.
+
+        SELECT account_balance("current.usd")
+
+    Supported core functions:
+        get_version
+        asset_list
+        asset_list_rates
+        asset_rate
+        asset_rate_range
+        account_info
+        account_statement
+        account_list
+        account_balance
+        account_balance_range
+
+    Args:
+        q: query to execute
+
+    Returns:
+        List of dicts is always returned
+    Raises:
+        RuntimeError: unsupported statement / function called
+        other: passed from called function as-is
+    """
+    q = q.strip()
+    if q[:7].lower() != 'select ':
+        raise RuntimeError('Unsupported statement')
+    try:
+        fn, args, kwargs = parse_func_str(q.split(maxsplit=1)[1])
+    except Exception as e:
+        raise ValueError(f'Invalid function params: {e}')
+    if fn == 'get_version':
+        yield {'variable': 'version', 'value': get_version(*args, **kwargs)}
+    elif fn == 'asset_list':
+        for row in asset_list(*args, **kwargs):
+            yield row
+    elif fn == 'asset_list_rates':
+        for row in asset_list_rates(*args, _grafana=_grafana, **kwargs):
+            yield row
+    elif fn == 'asset_rate':
+        result = asset_rate(*args,
+                            _grafana=_grafana,
+                            return_pair=True,
+                            **kwargs)
+        yield {'pair': result[0], 'rate': result[1]}
+    elif fn == 'account_info':
+        result = account_info(*args, **kwargs)
+        if isinstance(result, dict):
+            yield result
+        else:
+            for row in result:
+                yield row
+    elif fn == 'account_statement':
+        for row in account_statement(*args, _grafana=_grafana, **kwargs):
+            yield row
+    elif fn == 'account_list':
+        for row in account_list(*args, _grafana=_grafana, **kwargs):
+            yield row
+    elif fn == 'account_balance':
+        yield {'balance': account_balance(*args, _grafana=_grafana, **kwargs)}
+    elif fn == 'account_balance_range':
+        times, data = account_balance_range(*args, _grafana=_grafana, **kwargs)
+        for t, d in zip(times, data):
+            yield {'date': t, 'balance': d}
+    elif fn == 'asset_rate_range':
+        times, data = asset_rate_range(*args, _grafana=_grafana, **kwargs)
+        for t, d in zip(times, data):
+            yield {'date': t, 'rate': d}
+    else:
+        raise RuntimeError('Unsupported query function')
+
+
 @core_method
 def _asset_precision(asset=None):
     if asset:
-        if asset in asset_precision_cache:
-            return asset_precision_cache[asset]
+        if asset in _asset_precision_cache:
+            return _asset_precision_cache[asset]
         d = get_db().execute(sql('select precs from asset where code=:code'),
                              code=asset.upper()).fetchone()
         if not d:
             raise ResourceNotFound
         precs = int(d.precs)
-        asset_precision_cache[asset] = precs
+        _asset_precision_cache[asset] = precs
         return precs
     else:
 
@@ -256,7 +362,8 @@ def _asset_precision(asset=None):
                 sql('select code, precs from asset order by code'))
             while True:
                 r = d.fetchone()
-                if not r: break
+                if not r:
+                    break
                 row = OrderedDict()
                 row['asset'] = r.code
                 row['precision'] = int(r.precs)
@@ -323,7 +430,10 @@ _exceptions = {
     -32002: RateNotFound,
     -32003: OverdraftError,
     -32004: OverlimitError,
-    -32005: ResourceAlreadyExists
+    -32005: ResourceAlreadyExists,
+    -32601: AttributeError,
+    -32602: TypeError,
+    -32603: ValueError
 }
 
 
@@ -361,8 +471,10 @@ class AccountLocker:
 
     def release(self, token):
         with self._lock:
-            if token != self.token: raise RuntimeError('Invalid token')
-            if self.counter < 1: raise RuntimeError('Resource not locked')
+            if token != self.token:
+                raise RuntimeError('Invalid token')
+            if self.counter < 1:
+                raise RuntimeError('Resource not locked')
             self.counter -= 1
             if not self.counter:
                 if _db.redis_conn:
@@ -405,6 +517,10 @@ def get_db():
     return g.db
 
 
+def spawn(*args, **kwargs):
+    return _d.pool.submit(*args, **kwargs)
+
+
 def init(db=None, **kwargs):
     """
     Initialize finac database and configuration
@@ -412,6 +528,7 @@ def init(db=None, **kwargs):
     Args:
         db: SQLAlchemy DB URI or sqlite file name
         db_pool_size: DB pool size (default: 10)
+        thread_pool_size: thread pool size for internal processes (default: 30)
         keep_integrity: finac should keep database integrity (lock accounts,
             watch overdrafts, overlimits etc. Default is True
         lazy_exchange: allow direct exchange operations betwen accounts.
@@ -420,7 +537,8 @@ def init(db=None, **kwargs):
             "EUR/USD" pair exists but no USD/EUR, use 1 / "EUR/USD"
         rate_allow_cross: if exchange rate is not found, allow finac to look
             for the nearest cross-asset rate
-        rate_ttl: set rate cache ttl (default: 5 sec)
+        rate_cache_size: set rate cache size (default: 1024)
+        rate_cache_ttl: set rate cache ttl (default: 5 sec)
         full_transaction_update: allow updating transaction date and amount
         base_asset: default base asset. Default is "USD"
         date_format: default date format in statements
@@ -435,19 +553,26 @@ def init(db=None, **kwargs):
     Note: if Redis server is specified, Finac will use it for integrity locking
           (if enabled). In this case, lock tokens become Redis lock objects.
     """
+    rate_cache_ttl = 5
+    rate_cache_size = 1024
     for k, v in kwargs.items():
-        if k == 'rate_ttl':
-            rate_cache.__setstate__({'_TTLCache__ttl': v})
-        else:
-            if not hasattr(config, k):
-                raise RuntimeError('Parameter {} is invalid'.format(k))
+        if k == 'rate_cache_ttl':
+            rate_cache_ttl = v
+        elif k == 'rate_cache_size':
+            rate_cache_size = v
+        elif not hasattr(config, k):
+            raise RuntimeError('Parameter {} is invalid'.format(k))
         setattr(config, k, v)
+    _cache.rate = TTLCache(maxsize=rate_cache_size, ttl=rate_cache_ttl)
+    _cache.rate_list = TTLCache(maxsize=rate_cache_size, ttl=rate_cache_ttl)
+    config.rate_cache_ttl = rate_cache_ttl
+    _d.pool = ThreadPoolExecutor(max_workers=config.thread_pool_size)
     if config.multiplier:
         config.multiplier = float(config.multiplier)
     if db is not None:
         config.db = db
         db_uri = db
-        if db_uri.find('.db') != -1:
+        if db_uri.find('://') == -1:
             db_uri = 'sqlite:///' + os.path.expanduser(db_uri)
         _db.engine = get_db_engine(db_uri)
         _db.use_lastrowid = db_uri.startswith('sqlite') or db_uri.startswith(
@@ -461,6 +586,24 @@ def init(db=None, **kwargs):
                                      socket_timeout=config.redis_timeout)
 
 
+@core_method
+def config_set(prop, value):
+    """
+    Set configuration property on-the-fly (config.insecure = True required)
+
+    Recommended to use for testing only
+        
+    Args:
+        prop: property name
+        value: property value
+    """
+    if not config.insecure:
+        raise RuntimeError('access denied')
+    if not hasattr(config, prop):
+        raise RuntimeError('property {} is invalid'.format(prop))
+    setattr(config, prop, value)
+
+
 @core_method
 def asset_create(asset, precision=2):
     """
@@ -496,7 +639,8 @@ def asset_list():
         select code, precs from asset order by code"""))
     while True:
         d = r.fetchone()
-        if not d: break
+        if not d:
+            break
         row = OrderedDict()
         row['asset'] = d.code
         row['precision'] = d.precs
@@ -504,7 +648,11 @@ def asset_list():
 
 
 @core_method
-def asset_list_rates(asset=None, start=None, end=None):
+def asset_list_rates(asset=None,
+                     start=None,
+                     end=None,
+                     datefmt=False,
+                     _grafana=False):
     """
     List asset rates
 
@@ -513,6 +661,11 @@ def asset_list_rates(asset=None, start=None, end=None):
     If asset is not specified, "end" is used as date to get rates for all
     assets
     """
+    if _grafana:
+        if isinstance(start, int):
+            start = start / 1000
+        if isinstance(end, int):
+            end = end / 1000
     if asset:
         cond = ''
         asset = _safe_format(asset.upper())
@@ -539,7 +692,7 @@ def asset_list_rates(asset=None, start=None, end=None):
             from asset_rate
                 join asset as cf on asset_from_id = cf.id
                 join asset as ct on asset_to_id = ct.id
-                    where {cond}
+                    where {cond} order by d
         """.format(cond=cond)))
     else:
         d = parse_date(end, return_timestamp=False) if end else parse_date(
@@ -566,17 +719,18 @@ def asset_list_rates(asset=None, start=None, end=None):
                              d=d)
     while True:
         d = r.fetchone()
-        if not d: break
+        if not d:
+            break
         row = OrderedDict()
         row['asset_from'] = d.asset_from
         row['asset_to'] = d.asset_to
-        row['date'] = format_date(d.d)
+        row['date'] = format_date(d.d, force=datefmt)
         row['value'] = _demultiply(d.value)
         yield row
 
 
-@core_method
 @deletion_method
+@core_method
 def asset_delete(asset):
     """
     Delete asset
@@ -636,8 +790,8 @@ def asset_set_rate(asset_from, asset_to=None, value=None, date=None):
                      value=_multiply(value))
 
 
-@core_method
 @deletion_method
+@core_method
 def asset_delete_rate(asset_from, asset_to=None, date=None):
     """
     Delete currrency rate
@@ -664,8 +818,29 @@ def asset_delete_rate(asset_from, asset_to=None, date=None):
         raise ResourceNotFound
 
 
+def _parse_asset_pair(asset_from, asset_to):
+    if asset_from.find('/') != -1 and asset_to is None:
+        asset_from, asset_to = asset_from.split('/')
+    asset_from = asset_from.upper()
+    asset_to = asset_to.upper()
+    return asset_from, asset_to
+
+
 @core_method
-def asset_rate(asset_from, asset_to=None, date=None):
+def asset_rate(asset_from,
+               asset_to=None,
+               date=None,
+               _grafana=False,
+               return_pair=False):
+    asset_from, asset_to = _parse_asset_pair(asset_from, asset_to)
+    result = _asset_rate_lookup(asset_from,
+                                asset_to,
+                                date=date,
+                                _grafana=_grafana)
+    return (f'{asset_from}/{asset_to}', result) if return_pair else result
+
+
+def _asset_rate_lookup(asset_from, asset_to=None, date=None, _grafana=False):
     """
     Get asset rate for the specified date
 
@@ -673,22 +848,23 @@ def asset_rate(asset_from, asset_to=None, date=None):
 
     Function can be also called as e.g. asset_rate('EUR/USD')
     """
+    if _grafana:
+        if isinstance(date, int):
+            date = date / 1000
     if date is None:
         date = parse_date(return_timestamp=False)
     else:
         date = parse_date(date, return_timestamp=False)
-    if asset_from.find('/') != -1 and asset_to is None:
-        asset_from, asset_to = asset_from.split('/')
-    asset_from = asset_from.upper()
-    asset_to = asset_to.upper()
+    asset_from, asset_to = _parse_asset_pair(asset_from, asset_to)
     if asset_from == asset_to:
         return 1
     db = get_db()
 
-    def _get_rate(cf, ct):
+    def _get_rate(cf, ct, d):
+        key = _format_ttlcache_key(d, config.rate_cache_ttl)
         try:
-            return rate_cache[(cf, ct)]
-        except (KeyError, TypeError) as e:
+            return _cache.rate[(cf, ct, key)]
+        except _CacheRateKeyError:
             r = db.execute(sql("""
                 select value from asset_rate
                     join asset as cfrom on asset_from_id=cfrom.id
@@ -702,11 +878,10 @@ def asset_rate(asset_from, asset_to=None, date=None):
             d = r.fetchone()
             if d:
                 value = _demultiply(d.value)
-                try:
-                    rate_cache[(cf, ct)] = value
-                except TypeError:
-                    pass
+                _cache.rate[(cf, ct, key)] = value
                 return value
+            else:
+                return None
 
     def _get_crossrate(asset_from, asset_to, d):
 
@@ -722,14 +897,12 @@ def asset_rate(asset_from, asset_to=None, date=None):
 
         graph = {}
         rates = {}
+        key = _format_ttlcache_key(d, config.rate_cache_ttl)
         try:
-            ratelist = rate_cache[d]
-        except (KeyError, TypeError) as e:
+            ratelist = _cache.rate_list[key]
+        except _CacheRateListKeyError:
             ratelist = list(asset_list_rates(end=d))
-            try:
-                rate_cache[d] = ratelist
-            except TypeError:
-                pass
+            _cache.rate_list[key] = ratelist
         for r in ratelist:
             rates[(r['asset_from'], r['asset_to'])] = r['value']
             graph.setdefault(r['asset_from'], []).append(r['asset_to'])
@@ -737,27 +910,28 @@ def asset_rate(asset_from, asset_to=None, date=None):
             if (k[1], k[0]) not in rates:
                 rates[(k[1], k[0])] = 1 / v
                 graph.setdefault(k[1], []).append(k[0])
-        path = min(_find_path(graph, asset_from, asset_to), key=len)
+        try:
+            path = min(_find_path(graph, asset_from, asset_to), key=len)
+        except (KeyError, ValueError):
+            return None
         if not path:
-            return
+            return None
         rate = 1
         for i in range(0, len(path) - 1):
             rate *= rates[path[i], path[i + 1]]
         return rate
 
-    value = _get_rate(asset_from, asset_to)
+    value = _get_rate(asset_from, asset_to, date)
     if not value:
         if config.rate_allow_reverse is True:
-            value = _get_rate(asset_to, asset_from)
+            value = _get_rate(asset_to, asset_from, date)
             if not value:
                 if config.rate_allow_cross:
-                    try:
-                        value = _get_crossrate(asset_from, asset_to, date)
-                    except:
-                        pass
-                    if value: return value
-                raise RateNotFound('{}/{} for {}'.format(
-                    asset_from, asset_to, format_date(date)))
+                    value = _get_crossrate(asset_from, asset_to, date)
+                    if value:
+                        return value
+                raise RateNotFound('{}/{} for {} (base asset: {})'.format(
+                    asset_from, asset_to, format_date(date), config.base_asset))
             value = 1 / value
         else:
             raise RateNotFound
@@ -819,17 +993,23 @@ def account_create(account,
                        max_balance=_multiply(max_balance))
         acc_id = r.lastrowid if _db.use_lastrowid else r.fetchone().id
         db.execute(sql("""
-            insert into transact(account_debit_id, amount, d_created, d) values
-            (:account_id, 0, :d, :d)
+            INSERT INTO transact
+                (account_debit_id, amount, d_created, d, service)
+            VALUES
+                (:account_id, 0, :d, :d, :s)
             """),
                    account_id=acc_id,
-                   d=datetime.datetime(1970, 1, 1))
+                   d=datetime.datetime.fromtimestamp(0),
+                   s=True)
         db.execute(sql("""
-            insert into transact(account_credit_id, amount, d_created, d) values
-            (:account_id, 0, :d, :d)
+            INSERT INTO transact
+                (account_credit_id, amount, d_created, d, service)
+            VALUES
+                (:account_id, 0, :d, :d, :s)
             """),
                    account_id=acc_id,
-                   d=datetime.datetime(1970, 1, 1))
+                   d=datetime.datetime.fromtimestamp(0),
+                   s=True)
         dbt.commit()
     except IntegrityError:
         dbt.rollback()
@@ -841,29 +1021,54 @@ def account_create(account,
 
 
 @core_method
-def account_info(account):
+def account_info(account=None):
     """
     Get dict with account info
+
+    If no account is specified, generator object with info for all accounts is
+    returned
     """
-    r = get_db().execute(sql("""
+
+    def _format_qdata(d):
+        return {
+            'code': d.account_code,
+            'note': d.note,
+            'type': ACCOUNT_TYPE_NAMES[d.tp],
+            'tp': d.tp,
+            'passive': d.passive,
+            'asset': d.asset,
+            'max_overdraft': _demultiply(d.max_overdraft),
+            'max_balance': _demultiply(d.max_balance)
+        }
+
+    xkw = {}
+    if account is not None:
+        xkw['account'] = account.upper()
+    r = get_db().execute(
+        sql("""
             select account.code as account_code, account.note, account.tp,
             account.passive, asset.code as asset, max_overdraft, max_balance
             from account join
             asset on account.asset_id = asset.id
-            where account.code = :account"""),
-                         account=account.upper())
-    d = r.fetchone()
-    if not d: raise ResourceNotFound
-    return {
-        'code': d.account_code,
-        'note': d.note,
-        'type': ACCOUNT_TYPE_NAMES[d.tp],
-        'tp': d.tp,
-        'passive': d.passive,
-        'asset': d.asset,
-        'max_overdraft': _demultiply(d.max_overdraft),
-        'max_balance': _demultiply(d.max_balance)
-    }
+            {}""".format(
+            'where account.code = :account' if account is not None else '')),
+        **xkw)
+    if account is not None:
+        d = r.fetchone()
+        if not d:
+            raise ResourceNotFound
+        else:
+            return _format_qdata(d)
+    else:
+
+        def _listgen():
+            while True:
+                d = r.fetchone()
+                if not d:
+                    break
+                yield _format_qdata(d)
+
+        return _listgen()
 
 
 @core_method
@@ -876,6 +1081,7 @@ def transaction_info(transaction_id):
             transact.note as note,
             transact.d_created as d_created,
             transact.d as d,
+            transact.service as service,
             chain_transact_id,
             dt.code as debit,
             ct.code as credit
@@ -888,7 +1094,8 @@ def transaction_info(transaction_id):
             where transact.id = :transaction_id"""),
                          transaction_id=transaction_id)
     d = r.fetchone()
-    if not d: raise ResourceNotFound
+    if not d:
+        raise ResourceNotFound
     return {
         'id': transaction_id,
         'amount': _demultiply(d.amount),
@@ -897,6 +1104,7 @@ def transaction_info(transaction_id):
         'created': d.d_created,
         'completed': d.d,
         'chain_transact_id': d.chain_transact_id,
+        'service': d.service,
         'dt': d.debit if hasattr(d, 'debit') else None,
         'ct': d.credit if hasattr(d, 'credit') else None,
     }
@@ -939,8 +1147,8 @@ def transaction_apply(fname):
     return result
 
 
-@core_method
 @deletion_method
+@core_method
 def account_delete(account, lock_token=None):
     """
     Delete account
@@ -1005,7 +1213,8 @@ def account_unlock(account, token):
     if config.keep_integrity:
         with lock_account_token:
             l = account_lockers.get(account.upper())
-        if not l: raise ResourceNotFound
+        if not l:
+            raise ResourceNotFound
         return l.release(token)
 
 
@@ -1212,7 +1421,8 @@ def _transaction_move(dt=None,
         if _ct_info is not None and _dt_info is not None and _ct_info[
                 'passive'] and _dt_info['passive']:
             ct, dt = dt, ct
-    if amount == 0: return
+    if amount == 0:
+        return
     if amount is not None and amount < 0:
         raise ValueError('Amount should be greater than zero')
     if config.keep_integrity:
@@ -1374,12 +1584,15 @@ def transaction_move(dt=None,
                                      _ct_info=ct_info,
                                      _dt_info=dt_info)
     finally:
-        if ctoken: account_unlock(ct, ctoken)
-        if dtoken: account_unlock(dt, dtoken)
+        if ctoken:
+            account_unlock(ct, ctoken)
+        if dtoken:
+            account_unlock(dt, dtoken)
 
 
 @core_method
-def transaction_complete(transaction_ids, completion_date=None,
+def transaction_complete(transaction_ids,
+                         completion_date=None,
                          lock_token=None):
     """
     Args:
@@ -1421,8 +1634,8 @@ def transaction_complete(transaction_ids, completion_date=None,
                     account_unlock(dt, token)
 
 
-@core_method
 @deletion_method
+@core_method
 def transaction_delete(transaction_ids):
     """
     Delete (mark deleted) transaction
@@ -1433,8 +1646,9 @@ def transaction_delete(transaction_ids):
     for transaction_id in ids:
         tinfo = transaction_info(transaction_id)
         if not get_db().execute(sql("""
-        update transact set
-        deleted=:ts where id=:id or chain_transact_id=:id"""),
+        UPDATE transact SET
+            deleted=:ts WHERE (id=:id or chain_transact_id=:id) AND
+            service IS null"""),
                                 ts=parse_date(return_timestamp=False),
                                 id=transaction_id).rowcount:
             logging.error('Transaction {} not found'.format(transaction_id))
@@ -1471,7 +1685,8 @@ def transaction_purge(_lock=True):
             dbt.rollback()
             raise
     finally:
-        if _lock: lock_purge.release()
+        if _lock:
+            lock_purge.release()
 
 
 @core_method
@@ -1545,7 +1760,13 @@ def transaction_copy(transaction_ids,
 
 
 @core_method
-def account_statement(account, start=None, end=None, tag=None, pending=True):
+def account_statement(account,
+                      start=None,
+                      end=None,
+                      tag=None,
+                      pending=True,
+                      datefmt=False,
+                      _grafana=False):
     """
     Args:
         account: account code
@@ -1553,11 +1774,17 @@ def account_statement(account, start=None, end=None, tag=None, pending=True):
         end: statement end date/time
         tag: filter transactions by tag
         pending: include pending transactions
+        datefmt: format date according to configuration
     Returns:
         generator object
     """
+    if _grafana:
+        if isinstance(start, int):
+            start = start / 1000
+        if isinstance(end, int):
+            end = end / 1000
     acc_info = account_info(account)
-    cond = 'transact.deleted is null and d_created > \'1970-01-03\''
+    cond = 'transact.deleted is null and transact.service is null'
     d_field = 'd_created' if pending else 'd'
     if start:
         dts = parse_date(start, return_timestamp=False)
@@ -1591,12 +1818,13 @@ def account_statement(account, start=None, end=None, tag=None, pending=True):
                          account=account.upper())
     while True:
         d = r.fetchone()
-        if not d: break
+        if not d:
+            break
         row = OrderedDict()
         for i in ('id', 'amount', 'cparty', 'tag', 'note'):
             row[i] = getattr(d, i)
-        row['created'] = format_date(d.d_created)
-        row['completed'] = format_date(d.d)
+        row['created'] = format_date(d.d_created, force=datefmt)
+        row['completed'] = format_date(d.d, force=datefmt)
         row['is_completed'] = d.d is not None
         row['amount'] = _demultiply(row['amount'])
         if acc_info['passive'] and row['amount']:
@@ -1609,7 +1837,8 @@ def account_statement_summary(account,
                               start=None,
                               end=None,
                               tag=None,
-                              pending=True):
+                              pending=True,
+                              datefmt=False):
     """
     Args:
         account: account code
@@ -1617,6 +1846,7 @@ def account_statement_summary(account,
         end: statement end date/time
         tag: filter transactions by tag
         pending: include pending transactions
+        datefmt: format date according to configuration
     Returns:
         dict with fields:
             debit: debit turnover
@@ -1629,7 +1859,8 @@ def account_statement_summary(account,
                           start=start,
                           end=end,
                           tag=tag,
-                          pending=pending))
+                          pending=pending,
+                          datefmt=datefmt))
     credit = 0
     debit = 0
     for row in statement:
@@ -1663,8 +1894,10 @@ def account_list(asset=None,
                  passive=None,
                  code=None,
                  date=None,
+                 base=None,
                  order_by=['tp', 'asset', 'account', 'balance'],
-                 hide_empty=False):
+                 hide_empty=False,
+                 _grafana=False):
     """
     List accounts and their balances
 
@@ -1674,10 +1907,14 @@ def account_list(asset=None,
         passive: list passive, active or all (if None) accounts
         code: filter by acocunt code (may contain '%' as a wildcards)
         date: get balances for the specified date
+        base: convert account balances to base currency
         order_by: list ordering
         hide_empty: hide accounts with zero balance, default is False
     """
     cond = "transact.deleted is null"
+    if _grafana:
+        if isinstance(date, int):
+            date = date / 1000
     if tp:
         tp = _safe_format(tp)
         if not isinstance(tp, (list, tuple)):
@@ -1767,7 +2004,8 @@ def account_list(asset=None,
                        oby=('order by ' + oby) if oby else '')))
     while True:
         d = r.fetchone()
-        if not d: break
+        if not d:
+            break
         # if zero is not a "real zero" - consider x < 0.000001 is zero
         if hide_empty is False or (abs(d.balance) > 0.000001):
             row = OrderedDict()
@@ -1781,6 +2019,8 @@ def account_list(asset=None,
             if row['passive'] and row['balance']:
                 row['balance'] *= -1
             row['balance'] = _demultiply(row['balance'])
+            if base:
+                row['balance'] *= asset_rate(row['asset'], base, date=date)
             yield row
 
 
@@ -1983,7 +2223,8 @@ def _account_summary(balance_type,
             oby=('order by ' + oby) if oby else '')))
     while True:
         d = r.fetchone()
-        if not d: break
+        if not d:
+            break
         if hide_empty is False or d.balance:
             row = OrderedDict()
             for i in ('account', 'type', 'asset', balance_type + '_balance'):
@@ -1997,8 +2238,12 @@ def _account_summary(balance_type,
 
 
 @core_method
-def account_balance(account=None, tp=None, base=None, date=None,
-                    _natural=False):
+def account_balance(account=None,
+                    tp=None,
+                    base=None,
+                    date=None,
+                    _natural=False,
+                    _grafana=False):
     """
     Get account balance
 
@@ -2012,6 +2257,9 @@ def account_balance(account=None, tp=None, base=None, date=None,
         raise ValueError('Account and type can not be specified together')
     elif not account and not tp:
         tp = [k for k in ACCOUNT_TYPE_IDS if ACCOUNT_TYPE_IDS[k] <= 1000]
+    if _grafana:
+        if isinstance(date, int):
+            date = date / 1000
     cond = "transact.deleted is null"
     dts = parse_date(date, return_timestamp=False) if date else parse_date(
         return_timestamp=False)
@@ -2054,16 +2302,57 @@ def account_balance(account=None, tp=None, base=None, date=None,
     return balance
 
 
+@core_method
+def asset_rate_range(start,
+                     asset_from,
+                     asset_to=None,
+                     end=None,
+                     step=1,
+                     return_timestamp=False,
+                     _grafana=False):
+    """
+    Get list of asset rates for the specified time range
+
+    Returns:
+        tuple with time series list and corresponding asset rate
+    """
+    if _grafana:
+        if isinstance(start, int):
+            start = start / 1000
+        if isinstance(end, int):
+            end = end / 1000
+    times = []
+    data = []
+    dt = parse_date(start, return_timestamp=False)
+    end_date = parse_date(
+        end, return_timestamp=False) if end else datetime.datetime.now()
+    delta = datetime.timedelta(days=step)
+    last_record = False
+    while dt < end_date or not last_record:
+        if dt == end_date:
+            break
+        elif dt > end_date:
+            last_record = True
+            if _grafana:
+                dt = datetime.datetime.now()
+        times.append(dt.timestamp() if return_timestamp else dt)
+        b = _asset_rate_lookup(asset_from, asset_to, date=dt)
+        data.append(b)
+        dt += delta
+    return times, data
+
+
 @core_method
 def account_balance_range(start,
                           account=None,
                           tp=None,
                           end=None,
                           step=1,
-                          return_timestamp=True,
-                          base=None):
+                          return_timestamp=False,
+                          base=None,
+                          _grafana=False):
     """
-    Get list of account balances for the specified range
+    Get list of account balances for the specified time range
 
     Args:
         account: account code
@@ -2072,7 +2361,7 @@ def account_balance_range(start,
         end: end date/time, if not specified, current time is used
         step: list step in days
         return_timestamp: return dates as timestamps if True, otherwise as
-            datetime objects. Default is True
+            datetime objects. Default is False
 
     Returns:
         tuple with time series list and corresponding balance list
@@ -2081,13 +2370,17 @@ def account_balance_range(start,
         raise ValueError('Account and type can not be specified together')
     elif not account and not tp:
         tp = [k for k in ACCOUNT_TYPE_IDS if ACCOUNT_TYPE_IDS[k] <= 1000]
+    if _grafana:
+        if isinstance(start, int):
+            start = start / 1000
+        if isinstance(end, int):
+            end = end / 1000
     times = []
     data = []
     acc_info = {'account': account} if account else {'tp': tp}
     dt = parse_date(start, return_timestamp=False)
     end_date = parse_date(
-        end, return_timestamp=False
-    ) if end else datetime.datetime.now() + datetime.timedelta(days=1)
+        end, return_timestamp=False) if end else datetime.datetime.now()
     delta = datetime.timedelta(days=step)
     last_record = False
     while dt < end_date or not last_record:
@@ -2095,6 +2388,8 @@ def account_balance_range(start,
             break
         elif dt > end_date:
             last_record = True
+            if _grafana:
+                dt = datetime.datetime.now()
         times.append(dt.timestamp() if return_timestamp else dt)
         b = account_balance(**acc_info, base=base, date=dt)
         data.append(b)
